

#include <stdio.h>
#include "frame.h"
#include "vector.h"
#include "povproto.h"
#include "colour.h"
#include "objects.h"
#include "povray.h"
#include "ray.h"
#include "matrices.h"
#include "render.h"

#ifdef PostProcessPatch
#include "postproc.h"

static void FreePostProcessTasks(void);

static FILE* post_file;
static PP_DATA_LISTS Post_Lines;
static COLOUR **Input_Lines;
static COLOUR **Output_Lines;
static static_height;
static DBL **Weights;


#define ALLOCATE_GRID(name, t, h, w) {name = \
  (t**)POV_MALLOC((h) * sizeof(t*),"post process"); \
  for(y=0; y<h; y++) \
    name[y] = (t*)POV_MALLOC((w)*sizeof(t),"post process");}

#define FREE_GRID(name, h) {if(name){ \
  for(y=0; y<h; y++) \
    POV_FREE(name[y]); \
  POV_FREE(name); \
  name=NULL;}}

/*****************************************************************************

 FUNCTION

  void Initialize_PostProcessing(void)
    Initializes pointer so that we know if we need to free memory after an
    interruptoin

  Preconditions:
    DoPostProcess was not started without finishing

  Postconditions:
    various pointers are set to NULL

******************************************************************************/
void Initialize_PostProcessing(void)
{
  static_height = 0;
  Input_Lines=NULL;
  Output_Lines=NULL;
  post_file=NULL;

  /* only allocate what we need */
  Weights = NULL;
  Post_Lines.Depth = NULL;
  Post_Lines.Colour = NULL;
  Post_Lines.INormal =  NULL;
  Post_Lines.PNormal = NULL;
  Post_Lines.Iuv = NULL;
  Post_Lines.IPoint = NULL;
  Post_Lines.Object = NULL;
}

/*****************************************************************************

 FUNCTION

  void Deinitialize_PostProcessing(void)
    frees memory (if necessary), closes the post-data file (if necessary),
    and frees the linked-list of post-process tasks

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    If any of the pointer were allocated and freed, they should have been
    set to null so that we don't accidently free them again.

  Postconditions:
    all post-processing data is cleaned up, memory is freed, and pointers
    set to zero

******************************************************************************/
void Deinitialize_PostProcessing(void)
{
  int y;

  int height=static_height;

  FreePostProcessTasks();
  Close_Post_File();

  if (height)
  {
    /* free this stuff to handle interruptions */
    FREE_GRID(Input_Lines, height);
    FREE_GRID(Output_Lines, height);
    FREE_GRID(Post_Lines.Depth, height);
    FREE_GRID(Post_Lines.Colour, height);
    FREE_GRID(Post_Lines.IPoint, height);
    FREE_GRID(Post_Lines.INormal, height);
    FREE_GRID(Post_Lines.PNormal, height);
    FREE_GRID(Post_Lines.Iuv, height);
    FREE_GRID(Post_Lines.Object, height);
    FREE_GRID(Weights, height);
  }
}

/*****************************************************************************

 FUNCTION

  void Open_Post_File()
    opens the post-processing file (static post_file) for
    write

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    The file "static post_file" is not currently open.
    opts.postProcessFlags contains flags for all necessary post-data.
    If post processing will be done, opts.postProcess is not NULL.

  Postconditions:
    the post-processing file (static variable post_file) is open for write

******************************************************************************/
void Open_Post_File()
{
  if (post_file) fclose(post_file); /* just in case */

  if (!(opts.Options & DISKWRITE) ||  /* won't be able to read image */
      !(opts.postProcess) ||          /* no post processing tasks */
      !(opts.postProcessFlags))       /* no need for any extra data */
  {
    post_file = NULL;
    return;
  }

  if (opts.Options & CONTINUE_TRACE)
    post_file = fopen("postdata.tmp",APPEND_BINFILE_STRING);
  else
    post_file = fopen("postdata.tmp",WRITE_BINFILE_STRING);
}

/*****************************************************************************

 FUNCTION

  void Open_Post_Input()
    opens the post-processing file (static variable post_file) for
    read

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    The file "static post_file" is not currently open.
    opts.postProcessFlags contains flags for all necessary post-data.
    If post processing will be done, opts.postProcess is not NULL.

  Postconditions:
    the post-processing file (static variable post_file) is open for read

******************************************************************************/
void Open_Post_Input()
{
  if (post_file) fclose(post_file); /* just in case */

  if (!(opts.Options & DISKWRITE) ||  /* won't be able to read image */
      !(opts.postProcess) ||          /* no post processing tasks */
      !(opts.postProcessFlags))       /* no need for any extra data */
  {
    post_file = NULL;
    return;
  }

  post_file = fopen("postdata.tmp",READ_BINFILE_STRING);
}

/*****************************************************************************

 FUNCTION

  void Write_Post_Line(PP_DATA *Line, int width)
    writes a line of post-process data to static post_file

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    The file "static post_file" is open for write.
    opts.postProcessFlags contains flags for all necessary post-data.

    Line is a single-dim array that contains the post-process data for a
      single horizontal line of the image.
    width is the size of the array (the width of the image)

  Postconditions:
    necessary data (indicated by opts.postProcessFlags) is written to
    "static post_file"

******************************************************************************/
void Write_Post_Line(PP_DATA *Line, int width)
{
  int x;

  if (!post_file) return;

  /* only write the stuff we need */
  for(x=0; x<width; x++)
  {
    if (opts.postProcessFlags & PP_FLAG_DEPTH)
      fwrite(&Line[x].Depth,sizeof(DBL),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_COLOUR)
      fwrite(&Line[x].Colour,sizeof(COLOUR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_IPOINT)
      fwrite(&Line[x].IPoint,sizeof(VECTOR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_INORMAL)
      fwrite(&Line[x].INormal,sizeof(VECTOR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_PNORMAL)
      fwrite(&Line[x].PNormal,sizeof(VECTOR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_IUV)
      fwrite(&Line[x].Iuv,sizeof(UV_VECT),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_OBJECT)
      fwrite(&Line[x].Object,sizeof(int),1,post_file);
  }
}

/*****************************************************************************

 FUNCTION

  void Read_Post_Line(PP_DATA_LISTS Data, int y, int width)
    reads a line of post-process data from static post_file

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    The file "static post_file" is open for read.
    opts.postProcessFlags contains flags for all necessary post-data.
      These flags must be exactly the same flags that were set when the
      file was written, or the data will be corrupt.

    Data is a struct that contains pointers to grids (2-dim arrays)
      for receiving the post-process data
    width the width of the image (width of the 2-dim arrays)
    y indicates which line (in Line) the data is to be stored in

  Postconditions:
    necessary data (indicated by opts.postProcessFlags) is read from
    "static post_file" and 

******************************************************************************/
void Read_Post_Line(PP_DATA_LISTS Data, int y, int width)
{
  int size;
  int x;

  if (!post_file) return;

  /* only read the stuff we need */
  for(x=0; x<width; x++)
  {
    if (opts.postProcessFlags & PP_FLAG_DEPTH)
      size = fread(&Data.Depth[y][x],sizeof(DBL),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_COLOUR)
      size = fread(&Data.Colour[y][x],sizeof(COLOUR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_IPOINT)
      size = fread(&Data.IPoint[y][x],sizeof(VECTOR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_INORMAL)
      size = fread(&Data.INormal[y][x],sizeof(VECTOR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_PNORMAL)
      size = fread(&Data.PNormal[y][x],sizeof(VECTOR),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_IUV)
      size = fread(&Data.Iuv[y][x],sizeof(UV_VECT),1,post_file);
    if (opts.postProcessFlags & PP_FLAG_OBJECT)
      size = fread(&Data.Object[y][x],sizeof(int),1,post_file);
  }
}

/*****************************************************************************

 FUNCTION

  void Close_Post_File()
    closes the post-processing file (static post_file)

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    The file "static post_file" is open.

  Postconditions:
    the post-processing file (static variable post_file) is closed, and
    the variable is set to NULL

******************************************************************************/
void Close_Post_File()
{
  if (!(opts.Options & DISKWRITE)) return;
  if (!post_file) return;

  fclose(post_file);
  post_file = NULL;
}

/*****************************************************************************

 FUNCTION

  void Delete_Post_File()
    deletes the temporary post-processing file

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    The file "static post_file" is not open.

  Postconditions:
    the file "postdata.tmp" is deleted

******************************************************************************/
void Delete_Post_File()
{
  DELETE_FILE("postdata.tmp");
}

/*****************************************************************************

 FUNCTION

  void Assign_PP_Data(PP_DATA *pp_data, INTERSECTION *Inter, 
                      COLOUR Colour, DBL depth)
    assigns data to the PP_DATA struct

  Preconditions:
    Inter points to the intersection continaing data to be stored
    Colour is the true color of the pixel (before gamma-correct and clipping)
    depth is the depth of the trace ray

  Postconditions:
    the data in pp_data has been filled in from the data provided by the
      input variables

******************************************************************************/
void Assign_PP_Data(PP_DATA *pp_data, INTERSECTION *Inter, COLOUR Colour, DBL depth)
{
  Assign_Colour(pp_data->Colour, Colour);
  Assign_Vector(pp_data->IPoint, Inter->IPoint);
  Assign_Vector(pp_data->INormal ,Inter->INormal);
  Assign_UV_Vect(pp_data->Iuv ,Inter->Iuv);
  Assign_Vector(pp_data->PNormal ,Inter->PNormal);
  pp_data->Depth = depth;

  /* pp_data->Object = Inter->Object->Post_Process_Id; */
}

/*****************************************************************************

 FUNCTION

  void DoPostProcess(char* filename)
    Post-process the file "filename" using the data in "postdata.tmp"
    and the tasks in opts.postProcess

  Preconditions:
    Initialize_PostProcessing was called at some point prior to this.
    The file "static post_file" is not open.
    filename is the full path+filename for the output image
    Output_File_Handle (static) is the handle used to create that file
    opts.postProcess points to the start of a linked list of post tasks
    opts.postProcessFlags indicates what kind of data is necessary for all
      of the tasks in the linked list
    that data (and only that data) was saved in the file "postdata.tmp" 
      during ray-tracing (using Open/Write/Close with thos same flags set)

  Postconditions:
    the image is post-processed using the tasks in opts.postProcess

******************************************************************************/
void DoPostProcess(char* filename)
{
  int x,y,tmp;
  int width, height;
  PP_TASK *pp_task;
  COLOUR **temp_swap;

  if (!opts.postProcess) return;

  COOPERATE_0

  /* open image */
  Open_File(Output_File_Handle, filename, &width, &height, 0, READ_MODE);

  /* save the size of all of this data in case we are interrupted */
  static_height = height;

  /* allocate memory */
  ALLOCATE_GRID(Input_Lines, COLOUR, height, width);
  ALLOCATE_GRID(Output_Lines, COLOUR, height, width);

  /* only allocate what we need */
  Post_Lines.Depth = NULL;
  Post_Lines.Colour = NULL;
  Post_Lines.INormal =  NULL;
  Post_Lines.PNormal = NULL;
  Post_Lines.Iuv = NULL;
  Post_Lines.IPoint = NULL;
  Post_Lines.Object = NULL;

  if (opts.postProcessFlags & PP_FLAG_DEPTH)
    ALLOCATE_GRID(Post_Lines.Depth, DBL, height, width);
  if (opts.postProcessFlags & PP_FLAG_COLOUR)
    ALLOCATE_GRID(Post_Lines.Colour, COLOUR, height, width);
  if (opts.postProcessFlags & PP_FLAG_IPOINT)
    ALLOCATE_GRID(Post_Lines.IPoint, VECTOR, height, width);
  if (opts.postProcessFlags & PP_FLAG_INORMAL)
    ALLOCATE_GRID(Post_Lines.INormal, VECTOR, height, width);
  if (opts.postProcessFlags & PP_FLAG_PNORMAL)
    ALLOCATE_GRID(Post_Lines.PNormal, VECTOR, height, width);
  if (opts.postProcessFlags & PP_FLAG_IUV)
    ALLOCATE_GRID(Post_Lines.Iuv, UV_VECT, height, width);
  if (opts.postProcessFlags & PP_FLAG_OBJECT)
    ALLOCATE_GRID(Post_Lines.Object, int, height, width);

  /* read image */
  tmp = 0;
  for(y=0; y<height; y++)
  {
    COOPERATE_1
    Read_Line(Output_File_Handle, Input_Lines[y], &tmp);
  }
  Close_File(Output_File_Handle);

  COOPERATE_0

  /* open post-data for input */
  Open_Post_Input();
  for(y=0; y<height; y++)
  {
    COOPERATE_1
    Read_Post_Line(Post_Lines, y, width);
  }
  Close_Post_File();

  COOPERATE_0

  /* post-process the image */
  for(pp_task = (PP_TASK*)opts.postProcess; pp_task != NULL; pp_task=pp_task->next)
  {
    (*(pp_task->doPostProcess))(pp_task,Input_Lines,Output_Lines,Post_Lines,height,width);

    /* if not last node, then swap */
    if (pp_task->next)
    {
      temp_swap = Input_Lines;
      Input_Lines = Output_Lines;
      Output_Lines = temp_swap;
    }
  }

  /* the data to output is in Input_Lines, since it was swapped */

  COOPERATE_0

  /* write the image file and update the GUI display */

  Open_File(Output_File_Handle, filename, &width, &height, 0, WRITE_MODE);
  for(y=0; y<height; y++)
  {
    Write_Line(Output_File_Handle, Output_Lines[y], y);

    /* do GUI output */
    for(x=0; x<width; x++)
    {
      POV_ASSIGN_PIXEL (x, y, Output_Lines[y][x]);
      plot_pixel(x, y, Output_Lines[y][x]);
      COOPERATE_1
    }

    POV_WRITE_LINE (Output_Lines[y], y)
  }
  Close_File(Output_File_Handle);

  COOPERATE_0

  //FREE_GRID(Input_Lines, height);
  //FREE_GRID(Output_Lines, height);

  if(Input_Lines)
  {
    for(y=0; y<height; y++)
      POV_FREE(Input_Lines[y]);
     POV_FREE(Input_Lines);
     Input_Lines=NULL;
  }
  if(Output_Lines)
  {
    for(y=0; y<height; y++)
      POV_FREE(Output_Lines[y]);
     POV_FREE(Output_Lines);
     Output_Lines=NULL;
  }


  FREE_GRID(Post_Lines.Depth, height);
  FREE_GRID(Post_Lines.Colour, height);
  FREE_GRID(Post_Lines.IPoint, height);
  FREE_GRID(Post_Lines.INormal, height);
  FREE_GRID(Post_Lines.PNormal, height);
  FREE_GRID(Post_Lines.Iuv, height);
  FREE_GRID(Post_Lines.Object, height);

}

/*****************************************************************************

 FUNCTION

  static void FreePostProcessTasks(void)
    frees the linked list of tasks stored in opts.postProcess

  Preconditions:
    if opts.postProcess is not NULL, then it points to the start of
      a linked list that contains post-processing tasks
    the "next" pointer in the last node of the list is NULL

  Postconditions:
    the linked list is freed, and opts.postProcess is set to NULL

******************************************************************************/
static void FreePostProcessTasks(void)
{
  /* free the post-process list */
  PP_TASK *pp_task = (PP_TASK*)opts.postProcess;
  while(pp_task != NULL)
  {
    PP_TASK *pp_task2 = pp_task;
    pp_task=pp_task->next;
    POV_FREE(pp_task2);
  }
  opts.postProcess = NULL;
}


/* ============================================================ */
/*  Focal Blur post process */
/* ============================================================ */

PP_FOCAL_BLUR *createPostFocalBlur(void)
{
  PP_FOCAL_BLUR *blur;
  blur = (PP_FOCAL_BLUR *)POV_MALLOC(sizeof(PP_FOCAL_BLUR),"focal blur");
  blur->doPostProcess = doPostFocalBlur;
  blur->next = NULL;
  blur->fieldStart = 0;
  blur->fieldDepth = BOUND_HUGE;
  blur->maxPixelBlur = 0;
  return blur;
}

#define usegauss 1

void doPostFocalBlur(PP_TASK *pp_task, COLOUR **in, COLOUR **out, PP_DATA_LISTS data, int height, int width)
{
  PP_FOCAL_BLUR *blur;
  DBL depth, blur_amount, dist;
  DBL pixel_dist;
  DBL weight1, weight2, w;
  int x,y;
  int xa,xb,ya,yb;
  DBL start,mid;
  DBL field_depth;
  DBL depth2;
  int iblur;
  int x2,y2;

  Status_Info("\nApplying 'focal blur' effect: 0%%");

  blur = (PP_FOCAL_BLUR *)pp_task;

  start = blur->fieldStart;
  field_depth = blur->fieldDepth;
  mid = start+field_depth/2.0;

  /* 
    NOTE - WEIGHTS IS A STATIC VARIABLE TO AVOID MEMORY LEAKS IF THIS
    POST-PROCESS TASK IS INTERRUPTED
  */
  ALLOCATE_GRID(Weights, DBL, height, width);

  /* initialize weights and output data */
  for(y=0; y<height; y++)
  {
    for(x=0; x<width; x++)
    {
      Weights[y][x] = 0.0;
      out[y][x][0] = 
      out[y][x][1] = 
      out[y][x][2] = 0.0;
    }
  }


  for(y=0; y<height; y++)
  {
    Status_Info("\nApplying 'focal blur' effect: %d%%",(y+1)*100/height);
    COOPERATE_1

    for(x=0; x<width; x++)
    {
      depth = data.Depth[y][x];

#define FIND_BLUR(d) (max((sqrt(((d)-mid)*2.0/field_depth) - 1.0),0.0)*(width/640.0));

      if (depth>mid)
      {
        if (depth > mid+field_depth/2)
        {
          /*dist = depth-mid;*/
          /*blur_amount = sqrt(dist*2.0/field_depth) - 1.0;*/
          blur_amount = FIND_BLUR(depth);
        }
        else
          blur_amount = 0.0;
      }
      else
      {
        if (depth < start)
        {
          dist = mid - field_depth/2;
          blur_amount = dist/depth - 1.0;

          /* compensate for resolution, assuming square pixels */
          blur_amount *= (width/640.0);
        }
        else
          blur_amount = 0.0;
      }

      /* don't go below zero for any reason */
      blur_amount = max(blur_amount, 0.0);

      if(blur_amount < 0.1)  /* maybe 0.5 should be the threshold? */
      {
        out[y][x][0] += in[y][x][0];
        out[y][x][1] += in[y][x][1];
        out[y][x][2] += in[y][x][2];
        Weights[y][x] += 1.0;
      }
      else
      {
        iblur = (int)(blur_amount+1.0);
        /* maximum blur amount */
        iblur = max(min(blur->maxPixelBlur,iblur),0);
        xa = x-iblur; xb=x+iblur;
        ya = y-iblur; yb=y+iblur;

        if(depth>mid)
        {
          /* too far away */
          weight1=0.0;
          for (x2=xa; x2<=xb; x2++)
          {
            for (y2=ya; y2<=yb; y2++)
            {
              if (!(y2>=0 && y2<height && x2>=0 && x2<width)) continue;
              if( (x2!=x || y2!= y) && fabs(data.Depth[y2][x2]-mid)<field_depth/2 ) continue;

              pixel_dist = sqrt((DBL)(x-x2)*(x-x2)+(DBL)(y-y2)*(y-y2));

#if usegauss
              w = 1.0/(2*M_PI*blur_amount) * exp(-pixel_dist*pixel_dist / (2*blur_amount));
#else
              w = max(0.0, blur_amount-pixel_dist)/blur_amount;
#endif
              /* blend more if closer to same depth */
              depth2 = data.Depth[y2][x2];
              if (depth2 > mid+field_depth/2)
              {
                DBL blur_amount2 = FIND_BLUR(depth2);
                w *= min(min(blur_amount2,blur->maxPixelBlur) / min(blur_amount,blur->maxPixelBlur),1.0);
              }
              else w = 0.0;

              weight1 += w;
            }
          }

          for (x2=xa; x2<=xb; x2++)
          {
            for (y2=ya; y2<=yb; y2++)
            {
              if (!(y2>=0 && y2<height && x2>=0 && x2<width)) continue;
              if( (x2!=x || y2!= y) && fabs(data.Depth[y2][x2]-mid)<field_depth/2 ) continue;

              pixel_dist = sqrt((DBL)(x-x2)*(x-x2)+(DBL)(y-y2)*(y-y2));

#if usegauss
              w = 1.0/(2*M_PI*blur_amount) * exp(-pixel_dist*pixel_dist / (2*blur_amount));
              w /= weight1;
#else
              w = 
                max(0.0, blur_amount-pixel_dist) / (blur_amount * weight1);
#endif
              /* blend more if closer to same depth */
              depth2 = data.Depth[y2][x2];
              if (depth2 > mid+field_depth/2)
              {
                DBL blur_amount2 = FIND_BLUR(depth2);
                w *= min(min(blur_amount2,blur->maxPixelBlur) / min(blur_amount,blur->maxPixelBlur),1.0);
              }
              else w = 0.0;

              if (w==0.0) continue;

              out[y][x][0] += in[y2][x2][0]*w;
              out[y][x][1] += in[y2][x2][1]*w;
              out[y][x][2] += in[y2][x2][2]*w;
              Weights[y][x] += w;
            }
          }
        }
        else
        {
          /* too close */
          weight1=weight2=0.0;
          for (x2=xa; x2<=xb; x2++)
          {
            for (y2=ya; y2<=yb; y2++)
            {
              pixel_dist = sqrt((DBL)(x-x2)*(x-x2)+(DBL)(y-y2)*(y-y2));
#if usegauss
              w = 1.0/(2*M_PI*blur_amount) * exp(-pixel_dist*pixel_dist / (2*blur_amount));
#else
              w = max(0.0, blur_amount-pixel_dist)/blur_amount;
#endif

              weight2 += w;

              if (!(y2>=0 && y2<height && x2>=0 && x2<width)) continue;

              weight1 += w;
            }
          }

          for (x2=xa; x2<=xb; x2++)
          {
            for (y2=ya; y2<=yb; y2++)
            {
              if (!(y2>=0 && y2<height && x2>=0 && x2<width)) continue;
              pixel_dist = sqrt((DBL)(x-x2)*(x-x2)+(DBL)(y-y2)*(y-y2));
              
#if usegauss
              w = 1.0/(2*M_PI*blur_amount) * exp(-pixel_dist*pixel_dist / (2*blur_amount));
              w /= weight1;
#else
              w = max(0.0, blur_amount-pixel_dist) / (blur_amount*weight1);
#endif

              if (w==0.0) continue;
              out[y][x][0] += in[y2][x2][0]*w;
              out[y][x][1] += in[y2][x2][1]*w;
              out[y][x][2] += in[y2][x2][2]*w;
              Weights[y][x] += w;

#if usegauss
              w = 1.0/(2*M_PI*blur_amount) * exp(-pixel_dist*pixel_dist / (2*blur_amount));
              w /= weight2;
#else
              w = max(0.0, blur_amount-pixel_dist) / (weight2);
#endif
              out[y2][x2][0] += in[y][x][0]*w;
              out[y2][x2][1] += in[y][x][1]*w;
              out[y2][x2][2] += in[y][x][2]*w;
              Weights[y2][x2] += w;
            }
          }
        }
      }
    }
  }

  for(y=0; y<height; y++)
  {
    COOPERATE_1
    for(x=0; x<width; x++)
    {
      Scale_Colour(out[y][x],out[y][x],1.0/Weights[y][x]);
      out[y][x][3]=in[y][x][3];
      out[y][x][4]=in[y][x][4];
    }
  }

  FREE_GRID(Weights, height);
}

/* ============================================================ */
/*  Depth post process */
/* ============================================================ */

PP_DEPTH *createPostDepth(void)
{
  PP_DEPTH *depth;
  depth = (PP_DEPTH *)POV_MALLOC(sizeof(PP_DEPTH),"depth post process");
  depth->next = NULL;
  depth->doPostProcess = doPostDepth;
  depth->fieldDepth = BOUND_HUGE;
  depth->fieldStart = 0;
  return depth;
}

void doPostDepth(PP_TASK *pp_task, COLOUR **in, COLOUR **out, PP_DATA_LISTS data, int height, int width)
{
  int x,y;
  PP_DEPTH *pp_depth;

  Status_Info("\nApplying 'depth' effect: 0%%");

  pp_depth = (PP_DEPTH *)pp_task;

  for(y=0; y<height; y++)
  {
    Status_Info("\nApplying 'depth' effect: %d%%",(y+1)*100/height);
    COOPERATE_1
    for(x=0; x<width; x++)
    {
      /* for now, convert to depth */
      DBL depth;
      DBL start = pp_depth->fieldStart;
      DBL field_depth = pp_depth->fieldDepth;

      depth = Post_Lines.Depth[y][x];
      out[y][x][0] = min(1.0,max(0.0,1.0-((depth-start)/field_depth)));
      out[y][x][1] = min(1.0,max(0.0,1.0-((depth-start)/field_depth)));
      out[y][x][2] = min(1.0,max(0.0,1.0-((depth-start)/field_depth)));
      out[y][x][3]=in[y][x][3];
      out[y][x][4]=in[y][x][4];
    }
  }
}

/* ============================================================ */
/*  Soft Glow post process */
/* ============================================================ */
PP_SOFT_GLOW *createPostSoftGlow(void)
{
  PP_SOFT_GLOW *glow;
  glow = (PP_SOFT_GLOW *)POV_MALLOC(sizeof(PP_SOFT_GLOW),"soft glow");
  glow->next = NULL;
  glow->doPostProcess = doPostSoftGlow;
  glow->blurAmount = 3;
  glow->glowAmount = 1.0;
  return glow;
}

void doPostSoftGlow(PP_TASK *pp_task, COLOUR **in, COLOUR **out, PP_DATA_LISTS data, int height, int width)
{
  int x,y,x2,y2;
  int xa,xb,ya,yb;
  DBL pixel_dist2;
  PP_SOFT_GLOW *glow;
  DBL weight,w;
  int i;
  DBL **gauss_w;
  int iblur;

  Status_Info("\nApplying 'soft glow' effect: 0%%");

  glow = (PP_SOFT_GLOW *)pp_task;

  iblur = (int)glow->blurAmount;

  /* precompute gaussian weights, since they are used a lot */

  /* allocate the gaussian-blur weights */
  ALLOCATE_GRID(gauss_w, DBL, (iblur*2+1), (iblur*2+1));

  for(y=0; y<=iblur*2; y++)
    for(x=0; x<=iblur*2; x++)
    {
      pixel_dist2 = (DBL)(x-iblur)*(x-iblur)+(DBL)(y-iblur)*(y-iblur);
      gauss_w[y][x] = 1.0/(2*M_PI*glow->blurAmount) * exp(-pixel_dist2 / (2*glow->blurAmount));
    }


  /* now, post-process the image */

  for(y=0; y<height; y++)
  {
    Status_Info("\nApplying 'soft glow' effect: %d%%",(y+1)*100/height);
    COOPERATE_1
    for(x=0; x<width; x++)
    {
      /* do a gaussian blur first */
      xa = x-iblur; xb=x+iblur;
      ya = y-iblur; yb=y+iblur;

      weight = 0.0;
      out[y][x][0] = out[y][x][1] = out[y][x][2] = 0.0;

      for (x2=xa; x2<=xb; x2++)
        for (y2=ya; y2<=yb; y2++)
        {
          if (!(y2>=0 && y2<height && x2>=0 && x2<width)) continue;

          w = gauss_w[y2-y+iblur][x2-x+iblur];
          out[y][x][0] += in[y2][x2][0]*w;
          out[y][x][1] += in[y2][x2][1]*w;
          out[y][x][2] += in[y2][x2][2]*w;
          weight += w;
        }
      Scale_Colour(out[y][x],out[y][x],1.0/weight);

      if(glow->glowAmount>0.0)
      {
        /* now, combine with original using LIGHTEN combination */
        out[y][x][0] *= (glow->glowAmount);
        out[y][x][1] *= (glow->glowAmount);
        out[y][x][2] *= (glow->glowAmount);

        for(i=0; i<3; i++)
        {
          if (in[y][x][i]>out[y][x][i])  /* input is brighter - keep it */
            out[y][x][i]=in[y][x][i];
          else if (glow->glowAmount < 1.0)  /* must blend */
            out[y][x][i] = out[y][x][i] + in[y][x][i] * (1.0-glow->glowAmount);
        }
      }
      else
      {
        /* now, combine with original using DARKEN combination */
        out[y][x][0] *= -(glow->glowAmount);
        out[y][x][1] *= -(glow->glowAmount);
        out[y][x][2] *= -(glow->glowAmount);

        for(i=0; i<3; i++)
        {
          if (in[y][x][i]<out[y][x][i])  /* input is darker - keep it */
            out[y][x][i]=in[y][x][i];
          else if (glow->glowAmount > -1.0)  /* must blend */
            out[y][x][i] = out[y][x][i] + in[y][x][i] * (1.0+glow->glowAmount);
        }
      }

      out[y][x][3]=in[y][x][3];
      out[y][x][4]=in[y][x][4];

      Clip_Colour(out[y][x], out[y][x]);
    }
  }

  FREE_GRID(gauss_w, (iblur*2+1));
}

#ifdef PostCartoonPatch

/* ============================================================ */
/*  Cartoonize post process */
/* ============================================================ */

PP_CARTOON *createPostCartoon(void)
{
  PP_CARTOON *cartoon;
  cartoon = (PP_CARTOON *)POV_MALLOC(sizeof(PP_CARTOON),"cartoon");
  cartoon->next = NULL;
  cartoon->doPostProcess = doPostCartoon;
  
  cartoon->blurAmount = 4.0;
  cartoon->hueThreshold = 0.1;
  cartoon->lumThreshold = 0.2;

  Make_Colour(cartoon->lineColour,0.0,0.0,0.0);
  cartoon->lineHueThreshold = 0.1;
  cartoon->lineWidth = 2.0;

  return cartoon;
}

void doPostCartoon(PP_TASK *pp_task, COLOUR **in, COLOUR **out, PP_DATA_LISTS data, int height, int width)
{
  int x,y,x2,y2;
  int xa,xb,ya,yb;
  DBL pixel_dist2;
  PP_CARTOON *cartoon;
  DBL weight,w;
  DBL **gauss_w;
  int iblur;
  DBL hue1,hue2,lum1,lum2;
  COLOUR edgeColor;
  DBL huediff;

  Status_Info("\nApplying 'cartoon' effect: 0%%");

  cartoon = (PP_CARTOON *)pp_task;

  iblur = (int)cartoon->blurAmount;

  /* precompute gaussian weights, since they are used a lot */

  /* allocate the gaussian-blur weights */
  ALLOCATE_GRID(gauss_w, DBL, (iblur*2+1), (iblur*2+1));

  for(y=0; y<=iblur*2; y++)
    for(x=0; x<=iblur*2; x++)
    {
      pixel_dist2 = (DBL)(x-iblur)*(x-iblur)+(DBL)(y-iblur)*(y-iblur);
      gauss_w[y][x] = 1.0/(2*M_PI*cartoon->blurAmount) * exp(-pixel_dist2 / (2*cartoon->blurAmount));
    }


  /* now, post-process the image */

  for(y=0; y<height; y++)
  {
    Status_Info("\nApplying 'cartoon' effect: %d%%",(y+1)*100/height);
    COOPERATE_1
    for(x=0; x<width; x++)
    {
      /* do a gaussian blur first */
      xa = x-iblur; xb=x+iblur;
      ya = y-iblur; yb=y+iblur;

      weight = 0.0;
      out[y][x][0] = out[y][x][1] = out[y][x][2] = 0.0;

      hue1 = RGBtoHue(in[y][x]);

      lum1 = GREY_SCALE(in[y][x]);

      for (x2=xa; x2<=xb; x2++)
        for (y2=ya; y2<=yb; y2++)
        {
          if (!(y2>=0 && y2<height && x2>=0 && x2<width)) continue;

          w = gauss_w[y2-y+iblur][x2-x+iblur];
          hue2 = RGBtoHue(in[y2][x2]);
          lum2 = GREY_SCALE(in[y2][x2]);
          
          /* blur only same hue */
          if (fabs(hue1-hue2)<cartoon->hueThreshold &&
              fabs(lum1-lum2)<cartoon->lumThreshold)
          {
            out[y][x][0] += in[y2][x2][0]*w;
            out[y][x][1] += in[y2][x2][1]*w;
            out[y][x][2] += in[y2][x2][2]*w;
            weight += w;
          }
        }
      Scale_Colour(out[y][x],out[y][x],1.0/weight);

      /* now, subtract a grey-scaled edge detection */
      xa = x-cartoon->lineWidth; xb=x+cartoon->lineWidth;
      ya = y-cartoon->lineWidth; yb=y+cartoon->lineWidth;

      w = -((cartoon->lineWidth*2+1)*(cartoon->lineWidth*2+1)-1);

      edgeColor[0] = edgeColor[1] = edgeColor[2] = 0.0;
      huediff = 0.0;
      weight = 0.0;

      for (x2=xa; x2<=xb; x2++)
        for (y2=ya; y2<=yb; y2++)
        {
          if (!(y2>=0 && y2<height && x2>=0 && x2<width)) continue;
          if (y2==y && x2==x)
          {
            edgeColor[0] += in[y2][x2][0]*w;
            edgeColor[1] += in[y2][x2][1]*w;
            edgeColor[2] += in[y2][x2][2]*w;
          }
          else
          {
            edgeColor[0] += in[y2][x2][0];
            edgeColor[1] += in[y2][x2][1];
            edgeColor[2] += in[y2][x2][2];

            hue2 = RGBtoHue(in[y2][x2]);
            huediff += fabs(hue1-hue2);
            weight++;
          }
        }
      /* need better greyscale! */
      if (huediff/weight > cartoon->lineHueThreshold)
      {
        /* clip edgeColor */
        edgeColor[0] = min(max(edgeColor[0],0.0),1.0);
        edgeColor[1] = min(max(edgeColor[1],0.0),1.0);
        edgeColor[2] = min(max(edgeColor[2],0.0),1.0);

        lum1 = GREY_SCALE(edgeColor);

        out[y][x][0]-=lum1;
        out[y][x][1]-=lum1;
        out[y][x][2]-=lum1;

        /* clip final color */
        out[y][x][0] = min(max(out[y][x][0],0.0),1.0);
        out[y][x][1] = min(max(out[y][x][1],0.0),1.0);
        out[y][x][2] = min(max(out[y][x][2],0.0),1.0);

        out[y][x][0]+=lum1*cartoon->lineColour[0];
        out[y][x][1]+=lum1*cartoon->lineColour[1];
        out[y][x][2]+=lum1*cartoon->lineColour[2];
      }

      out[y][x][3]=in[y][x][3];
      out[y][x][4]=in[y][x][4];

      /* clip final color */
      out[y][x][0] = min(max(out[y][x][0],0.0),1.0);
      out[y][x][1] = min(max(out[y][x][1],0.0),1.0);
      out[y][x][2] = min(max(out[y][x][2],0.0),1.0);

    }
  }

  FREE_GRID(gauss_w, (iblur*2+1));
}

#endif

#else
static char dummy[2];
#endif
